import abc
import asyncio
import hashlib
import os
import shutil
import yaml
from . import paths, wine, config


# Step
class Step(abc.ABC):
    @abc.abstractclassmethod
    async def run(self, wine, instance):
        raise NotImplementedError()

    @abc.abstractclassmethod
    async def content_to_hash(self, wine, instance):
        raise NotImplementedError()


class _StepPreInit(Step):
    def _find_drive_c(self, wine):
        return os.path.join(os.path.abspath(wine.path), "drive_c")

    def _gen_paths(self, wine, instance):
        # Find paths
        drive_c = self._find_drive_c(wine)
        volume_dir = os.path.abspath(os.path.join(paths.VOLUME_DIR, instance.get_instance()))

        # Iterate
        for volume, path in instance.get_volumes().items():
            if "\x00" in volume or "/" in volume or "\\" in volume:
                raise ValueError("Volume name contains slash or backslash")
            voldir = os.path.join(volume_dir, volume)
            pathdir = os.path.join(drive_c, path)
            yield (voldir, pathdir)

    async def run(self, wine:wine.WineConfig, instance):
        # Reset path
        shutil.rmtree(wine.path)
        os.makedirs(self._find_drive_c(wine), exist_ok=True)

        # Create volumes
        for volume, path in self._gen_paths(wine, instance):
            os.makedirs(volume, exist_ok=True)
            os.makedirs(os.path.dirname(path), exist_ok=True)
            os.symlink(os.path.relpath(volume, os.path.dirname(path)), path)

    async def content_to_hash(self, wine, instance):
        return b"\x00".join(map(lambda x: b"%s\x00%s" % x, map(lambda x: (x[0].encode("UTF-8"), x[1].encode("UTF-8")), sorted(self._gen_paths(wine, instance)))))


class _StepWineInit(Step):
    async def run(self, wine, isinstance):
        result = await wine.run_command([b"wineboot", b"--init"], with_display=False)
        if not result:
            raise RuntimeError() # TODO: gen exception + handling
        return result

    async def content_to_hash(self, wine, instance):
        version = await asyncio.create_subprocess_exec(b"wine", b"--version", stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
        stdout, stderr = await version.communicate()
        if version.returncode != 0:
            raise RuntimeError("Can't get wine version.")
        return stdout.splitlines()[0]


class _StepExec(Step):
    __command: list
    __in_drive_c: bool
    __cwd_path: str

    def __init__(self, command, in_drive_c: bool = False, cwd_path: str = None):
        self.__command = list(map(str.encode, command))
        self.__in_drive_c = in_drive_c
        self.__cwd_path = cwd_path

    async def run(self, wine, instance):
        tmp_cwd = self.__cwd_path
        if self.__in_drive_c:
            tmp_cwd = None
        result = await wine.run_command(args=self.__command, cwd_in_instance=self.__in_drive_c, cwd_path=tmp_cwd)
        if not result:
            print(repr(result))
            raise RuntimeError() # TODO: Exception
        return result

    async def content_to_hash(self, wine, isinstance):
        return b"\x00".join(self.__command)


# Instance management
class Instance():
    __instance: str
    __run: _StepExec
    __wine: wine.WineConfig
    __steps: list
    __volumes: dict
    __backup: config.ConfigAndBackup

    __has_to_restore_backup: bool

    def __gen_hash(self, last_hash:bytes, name:str, data:bytes):
        hasher = hashlib.sha3_256()
        hasher.update(last_hash)
        hasher.update(name.encode("UTF-8"))
        hasher.update(data)
        return hasher.hexdigest()

    def __init__(self, instance: str, run: _StepExec, steps: list = None, volumes: dict = {}):
        if not instance:
            raise ValueError("instance name can't be empty")
        self.__instance = str(instance)
        self.__run = run
        self.__wine = wine.WineConfig(os.path.join(
            paths.INSTANCE_DIR, instance), os.path.join(paths.FAKEHOME_DIR, instance))
        self.__steps = list(steps)
        self.__volumes = dict(volumes)
        self.__has_to_restore_backup = True

    async def __run_install(self, name: str, step: Step, last_hash: bytes, stop_wine:bool=True):
        if not isinstance(step, Step):
            raise ValueError("step have to be a Step.")
        # TODO: Check if allready done
        # TODO: Update gui (show install window)
        # TODO: Update gui (show that step is running)
        result = await step.run(self.__wine, self)
        if stop_wine:
            await self.__wine.stop_wine()
        
        # Gen backup and return
        next_hash = self.__gen_hash(last_hash, name, await step.content_to_hash(self.__wine, self))
        await self.__backup.gen_install_backup(next_hash)
        return next_hash.encode("UTF-8")

    async def install(self):
        # TODO: Set gui information
        last_hash = await self.__run_install("", _StepPreInit(), b"", stop_wine=False)
        last_hash = await self.__run_install("init", _StepWineInit(), last_hash)
        for name, step in self.__steps:
            last_hash = await self.__run_install(name, step, last_hash)

    def get_instance(self):
        return self.__instance
    
    def get_volumes(self):
        return dict(self.__volumes)
    
    def get_instance_path(self):
        return self.__wine.path
    
    async def is_installed(self):
        return False # TODO

    async def run(self):
        # Create backup
        self.__backup = config.ConfigAndBackup(self)
        await self.__backup.init()

        # Install and run
        await self.install()
        # TODO: Update gui (hide main window)
        await self.__run.run(self.__wine, self)


# Create instance
def _get_relative_file(current_file, file):
    return os.path.abspath(os.path.join(os.path.split(current_file)[0], file))


def _gen_run(file, config):
    config = dict(config)

    # Load command
    if "command" not in config:
        raise ValueError(
            "Can't find command in exec in file " + repr(file) + ".")
    if not isinstance(config["command"], list):
        raise ValueError(
            "Command in exec is not an array in file " + repr(file) + ".")
    command = config["command"]
    del config["command"]

    # Check if use in_drive_c
    in_drive_c = False
    if "in_drive_c" in config:
        if config["in_drive_c"]:
            in_drive_c = bool(config["in_drive_c"])
        del config["in_drive_c"]

    # Check for unused arguments
    if config:
        raise ValueError("Unused attributes " +
                         ", ".join(config.keys()) + " in file " + repr(file) + ".")

    # Return new exec step
    return _StepExec(command=command, in_drive_c=in_drive_c, cwd_path=os.path.dirname(file))


def _load_volumes(file, config):
    # Precheck
    if not isinstance(config, list):
        raise ValueError(
            "Volume list have to be a list in file " + repr(file) + ".")

    # Volumes
    volumes = []
    for i in config:
        # Check entry
        if not isinstance(i, dict):
            raise ValueError(
                "Volume instance have to be an dict in file " + repr(file) + ".")
        i = dict(i)

        # Load name and path
        if "name" not in i:
            raise ValueError(
                "Can't find name in volume info in file " + repr(file) + ".")
        if "path" not in i:
            raise ValueError(
                "Can't find path in volume info in file " + repr(file) + ".")
        name = str(i["name"])
        path = str(i["path"])
        del i["name"]
        del i["path"]
        volumes.append((name, path))

        # Check for unused attributes
        if i:
            raise ValueError("Volume has unused attirbutes " +
                             ", ".join(i.keys()) + " in file " + repr(file) + ".")
    return volumes


def _load_module(file: str, content: dict):
    # Check if string to load file
    if isinstance(content, str):
        file = _get_relative_file(file, content)
        with open(file, "rb") as f:
            content = yaml.load(f, Loader=yaml.SafeLoader)
    content = dict(content)

    # Load sub modules
    result = []
    volumes = []
    if "modules" in content:
        modules = content["modules"]
        del content["modules"]
        if not isinstance(modules, list):
            raise ValueError(
                "\"modules\" isn't a list in file " + repr(modules) + ".")
        for i in modules:
            tmp_result, tmp_volumes = _load_module(file, i)
            result += tmp_result
            volumes += tmp_volumes

    # Load volumes
    if "volumes" in content:
        volumes += _load_volumes(file, content["volumes"])
        del content["volumes"]

    # Load name
    if "name" not in content:
        raise ValueError(
            "name is missing in module in file " + repr(file) + ".")
    name = content["name"]
    del content["name"]

    # Check if type exists
    if "type" not in content:
        raise ValueError("Can't find type in file " + repr(file) + ".")

    type_name = content["type"]
    del content["type"]
    if type_name == "exec":
        if tuple(content.keys()) != ("run",):
            raise ValueError("Exec only supports \"run\" attribute but has " +
                             ", ".join(content.keys()) + " in file " + repr(file) + ".")
        result.append((name, _gen_run(file, content["run"])))
    else:
        raise ValueError("Type " + repr(type_name) +
                         " is unknown in file " + repr(file) + ".")

    # Cleanup modules
    return result, volumes


def gen_instance(file):
    # Load base data
    with open(file, "rb") as f:
        content = dict(yaml.load(f, Loader=yaml.SafeLoader))
    if "name" not in content:
        raise ValueError("Can't find name inside of " + repr(file) + ".")
    name = content["name"]
    del content["name"]

    # Load run
    if "run" not in content:
        raise ValueError("Can't find run inside of " + repr(file) + ".")
    run = _gen_run(file, content["run"])
    del content["run"]

    # Load modules
    modules = []
    volumes = []
    if "modules" not in content:
        raise ValueError("Can't find \"modules\" in " + repr(file) + ".")
    if not isinstance(content["modules"], list):
        raise ValueError(
            "\"modules\" have to be an array in file " + repr(file) + ".")
    for i in content["modules"]:
        tmp_modules, tmp_volumes = _load_module(file, i)
        modules += tmp_modules
        volumes += tmp_volumes
    del content["modules"]

    # Check unused attributes
    if content:
        raise ValueError("Unused attributes " +
                         ", ".join(content.keys()) + " in file " + repr(file) + ".")

    # Return new instance
    return Instance(instance=name, run=run, steps=modules, volumes=dict(volumes))
